import{r as o,e as d,k as l,R as i}from"./index.1f70e7cd.js";const a=(e="2.2.3",n=()=>{})=>{const t="__bokeh_lib_injection__",r=document.getElementById(t);if(r)window.Bokeh?n():r.addEventListener("load",n);else{const s=document.createElement("script");s.src=`https://cdn.bokeh.org/bokeh/release/bokeh-${e}.min.js`,s.id=t,document.body.appendChild(s),s.addEventListener("load",n)}},c=e=>{const[n,t]=o.exports.useState(!1);return e.bokehJson?.version?n?(window.Bokeh.logger.set_level("error"),o.exports.createElement(h,{...e})):(a(e.bokehJson.version,()=>{t(!0)}),o.exports.createElement(o.exports.Fragment,null)):o.exports.createElement(o.exports.Fragment,null,"-")},h=e=>{const n=o.exports.useRef(null),t=o.exports.useMemo(()=>"_bokeh__"+new Date().getTime(),[]);return o.exports.useEffect(()=>{if(e.bokehJson){n.current&&(n.current.innerHTML="");let r=0;e.bokehJson.roots.root_ids!=null&&(r=e.bokehJson.roots.root_ids[0]),window.Bokeh.embed.embed_item({doc:e.bokehJson,root_id:r,target_id:t})}},[e.bokehJson,t]),o.exports.createElement("div",{style:{padding:"10px",height:"100%",width:"100%"}},o.exports.createElement(d,{className:"bokeh-card",style:{width:"100%"}},e.headerElements,!!e.contentNotFound||!e.bokehJson?e.notFoundElements?e.notFoundElements:null:o.exports.createElement("div",{ref:n,id:t})))},m=e=>{const n=e.input,t=l(n);return t.loading?i.createElement("div",null):i.createElement(c,{bokehJson:t.contents?JSON.parse(t.contents):null})};export{m as default};
//# sourceMappingURL=Component.ecc5bf67.js.map
