import{u as n,R as t,P as r,S as a,m as u}from"./index.1f70e7cd.js";const o=l=>{const e=n(l.input);return e.loading?t.createElement(r,null):e.result==null?t.createElement(a,null,t.createElement(u,null)):t.createElement("div",{"data-test-weave-id":"boolean",style:{display:"flex",justifyContent:"center",alignItems:"center",height:"100%",width:"100%",textAlign:"center"}},t.createElement("div",{style:{lineHeight:"1em",padding:"0.3em 0.5em",borderRadius:"0.5em",backgroundColor:e.result?"#00A36815":"#FF667016",color:e.result?"#008F5D":"#EB1C45"}},e.result?"True":"False"))};export{o as default};
//# sourceMappingURL=Component.c0ed6440.js.map
