import{ah as i,u as o,R as e}from"./index.1f70e7cd.js";const n=l=>{const a=l.input,r=i({file:a}),t=o(r);return e.createElement("div",null,t.loading?e.createElement("div",null):e.createElement("img",{style:{maxWidth:"100%"},alt:"cool-alt",src:t.result}))};export{n as default};
//# sourceMappingURL=Component.d3ff73aa.js.map
